#' Diagnostic plot for the number of iterations used in sequential imputation methods
#' 
#' @description The \code{choosemaxit} function plots the mean of variables (specified in \code{plotvars}) against the iteration number for each of the replications (specified in \code{plotm}).
#' 
#' @param output an outpout from the imputedata function
#' @param plotvars index of variables for which a curve is plotted
#' @param cex a numerical vector giving the amount by which plotting characters and symbols should be scaled relative to the default
#' @param pch a vector of plotting characters or symbols
#' @param type what type of plot should be drawn
#' @param xlab a title for the x axis
#' @param ylab a title for the y axis
#' @param mar a numerical vector of the form c(bottom, left, top, right)
#' @param col a specification for the plotting color
#' @param mfrow a vector of the form c(nr, nc)
#' @param plotm a vector indicating which imputed datasets must be plotted
#' @export
#' @importFrom grDevices rgb
#' @examples
#' data(wine)
#' set.seed(123456)
#' wine.na <- wine
#' wine.na$cult <- NULL
#' wine.na <- as.matrix(wine.na)
#' wine.na[sample(seq(length(wine.na)), size = ceiling(length(wine.na)/3))] <- NA
#' nb.clust <- 3 # number of clusters
#' m <- 3 # number of imputed data sets
#' maxit <- 50 # number of iterations for FCS imputation
#' res.imp <- imputedata(data.na = wine.na, method = "FCS-homo",
#'                       nb.clust = nb.clust, m = m, maxit = maxit)
#' choosemaxit(res.imp, cex = .7)


choosemaxit<-function(output,plotvars=NULL,plotm=1:5,cex=.8,
                      pch=16,type="b",xlab="iterations",ylab="mean",
                      mar=c(4, 4, 2, 1) + 0.1,col=rgb(0,0,0,alpha=0.4),mfrow=NULL){
  if(is.null(plotvars)){indexvar<-seq(dim(output$res.conv)[3])}else{indexvar<-plotvars}
  p <- length(indexvar)
  Mfrow_old <- par()$mfrow
  if(is.null(mfrow)){
  Mfrow <- c(min(p, 4), 1 + (p - 1)%/%4)
  
  }else{
    Mfrow <- mfrow
  }
  plotm.intern<-plotm[which(plotm%in%seq(dim(output$res.conv)[1]))]
  par(mfrow = Mfrow,mar=mar)
  for (vtmp in indexvar){
    matplot(t(output$res.conv[plotm.intern,,vtmp]),
            main=colnames(output$call$data.na)[vtmp],
            cex=cex,pch=pch,type=type,ylab=ylab,xlab=xlab,
            col=col)
  }
  par(mfrow = Mfrow_old)
  # return()
}
