% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterMI.R
\name{clusterMI}
\alias{clusterMI}
\title{Cluster analysis and pooling after multiple imputation}
\usage{
clusterMI(
  output,
  method.clustering = "kmeans",
  method.consensus = "NMF",
  scaling = TRUE,
  nb.clust = NULL,
  Cboot = 50,
  threshold.saom = 200,
  method.agnes = "average",
  modelNames = NULL,
  nstart.kmeans = 100,
  m.cmeans = 2,
  nnodes = 1,
  instability = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{output}{an output from the imputedata function}

\item{method.clustering}{a single string specifying the clustering algorithm used ("kmeans", "pam", "agnes" or "mixture")}

\item{method.consensus}{a single string specifying the consensus method used to pool the contributory partitions}

\item{scaling}{boolean. If TRUE, variables are scaled. Default value is TRUE}

\item{nb.clust}{an integer specifying the number of clusters}

\item{Cboot}{an integer specifying the number of bootstrap replications. Default value is 50}

\item{threshold.saom}{a integer specifying when the algorithm is stopped}

\item{method.agnes}{character string defining the clustering method for hierarchical clustering (required only if method.clustering = "agnes")}

\item{modelNames}{character string indicating the models to be fitted in the EM phase of clustering (required only if method.clustering = "mixture"). By default modelNames = NULL.}

\item{nstart.kmeans}{how many random sets should be chosen for kmeans initalization. Default value is 100 (required only if method.clustering = "kmeans")}

\item{m.cmeans}{degree of fuzzification in cmeans clustering. By default m.cmeans = 2}

\item{nnodes}{number of CPU cores for parallel computing. By default, nnodes = 1}

\item{instability}{a boolean indicating if cluster instability must be computed. Default value is TRUE}

\item{verbose}{a boolean. If TRUE, a message is printed at each step. Default value is TRUE}
}
\description{
From a list of imputed datasets \code{clusterMI} performs cluster analysis on each imputed data set, estimates the instability of each partition using bootstrap (Fang, Y. and Wang, J., 2012) and pools results as proposed in Audigier and Niang (2020)
}
\details{
\code{clusterMI} performs cluster analysis (according to the \code{method.clustering} argument) and pooling after multiple imputation. For achieving this goal, the \code{clusterMI} function uses as an input an output from the \code{imputedata} function and then
\enumerate{
 \item applies the cluster analysis method on each imputed data set
 \item pools contributory partitions using non-negative matrix factorization
 \item computes the instability of each partition by bootstrap
 \item computes the total instability
 }
 
Step 1 can be tuned by specifying the cluster analysis method used (\code{method.clustering} argument).
If \code{method.clustering = "kmeans"}  or \code{"pam"}, the number of clusters can be specified by tuning the \code{nb.clust} argument. By default, the same number as the one used for imputation is used.
The number of random initializations can also be tuned through the \code{nstart.kmeans} argument.
If \code{method.clustering = "agnes"} (hierarchical clustering), the method used can be specified (see \code{\link[cluster]{agnes}}). By default \code{"average"} is used. Furthermore, the number of clusters can be specified, but it can also be automatically chosen if \code{nb.clust} < 0.
If \code{method.clustering = "mixture"} (model-based clustering using gaussian mixture models), the model to be fitted can be tuned by modifying the \code{modelNames} argument (see \code{\link[mclust]{Mclust}}).

Step 2 performs consensus clustering. The default method used is "NMF" (Li and Ding, 2007), which is recommended.
Alternatively, Simulated-Annealing-One-element-Move (Filkov and Skiena, 2004) can be used (\code{method.consensus = "Mirkin"}). This method starts from an intial consensus partition and looks at a "better" partition by stochastics steps.
If after \code{threshold.saom} iterations no improvements have been found, then the algorithm stops.

Step 3 applies the \code{\link[fpc]{nselectboot}} function on each imputed data set and returns the instability of each cluster obtained at step 1. The method is based on bootstrap sampling (Fang, Y. and Wang, J., 2012). The number of iterations can be tuned using the \code{Cboot} argument.

Step 4 averages the previous instability measures given a within instability (\code{Ubar}), computes a between instability (\code{B}) and a total instability (\code{T} = B + Ubar). See Audigier and Niang (2020) for details.

All steps can be performed in parallel by specifying the number of CPU cores (\code{nnodes} argument). Steps 3 and 4 are more time consuming. To compute only steps 1 and 2 use \code{instability = FALSE}.
}
\examples{
data(wine)

require(parallel)
set.seed(123456)
ref <- wine$cult
nb.clust <- 3
wine.na <- wine
wine.na$cult <- NULL
wine.na <- as.matrix(wine.na)
wine.na[sample(seq(length(wine.na)), size = ceiling(length(wine.na)/3))] <- NA

#imputation
res.imp <- imputedata(data.na = wine.na, nb.clust = nb.clust)

#pooling
nnodes <- 2 # parallel::detectCores()
res.pool <- clusterMI(res.imp, nnodes = nnodes)

res.pool$instability
table(ref, res.pool$part)
}
\references{
Audigier, V. and Niang, N. (2020) Clustering with missing data: which equivalent for Rubin's rules? ArXiv e-prints <arXiv:2011.13694>
  
  Fang, Y. and Wang, J. (2012) Selection of the number of clusters via the bootstrap method. Computational Statistics and Data Analysis, 56, 468-477. <doi:10.1016/j.csda.2011.09.003>
  
  Filkov and S. Skiena (2004) Integrating microarray data by consensus clustering. International Journal on Artificial Intelligence Tools, 13(04):863–880 <doi:10.1142/S0218213004001867>
  
  T. Li, C. Ding, and M. I. Jordan (2007) Solving consensus and semi-supervisedclustering problems using nonnegative matrix factorization.  InProceedingsof the 2007 Seventh IEEE International Conference on Data Mining, ICDM'07, page 577-582, USA. IEEE Computer Society. <doi:10.1109/ICDM.2007.98>
}
\seealso{
\code{\link[cluster]{agnes}}, \code{\link[fpc]{nselectboot}}, \code{\link[mclust]{Mclust}}, \code{\link{imputedata}}, \code{\link[e1071]{cmeans}}
}
